// COPYRIGHT 1994 A.D. Software, All rights reserved

// public layer of OOFILE database
// implementing relations

#include "oof3.hpp"
#include "oofrel.hpp"

bool dbRelationship::sAllJoinsUpdateJoinField = true;
 

// -------------------------------------------------------
//                   d b R e l a t i o n
// -------------------------------------------------------
dbRelationship::dbRelationship(dbRelRefBase& lhsRef, dbRelRefBase& rhsRef, const char* name) : 
												mlhs(new dbRelHalf(lhsRef)),
												mrhs(new dbRelHalf(rhsRef)),
												mName(name),
												mPartSides(neither),
												mJoinUpdatesJoinField(sAllJoinsUpdateJoinField)
{
	assert(dbConnect::underConstruction());   
// nil means this dbRel is being created AFTER the New or openConnection
// too late to be part of this database connection's schema
	
	dbConnect::sCurrentlyConstructing->attachRelation(this);
}

				
dbRelationship::dbRelationship(dbRelRefBase& lhsRef, dbRelRefBase& rhsRef, dbField& joinLeft, dbField& joinRight, const char* name) : 
												mlhs(new dbRelHalf(lhsRef)),
												mrhs(new dbRelHalf(rhsRef)),
												mName(name),
												mPartSides(neither),
												mJoinUpdatesJoinField(sAllJoinsUpdateJoinField)
{
	assert(dbConnect::underConstruction());   
	dbConnect::sCurrentlyConstructing->attachRelation(this);
	joinFields(joinLeft, joinRight);
}


dbRelationship:: dbRelationship(dbRelHalf* lhs, dbRelHalf* rhs) :
												mlhs(lhs),
												mrhs(rhs),
												mName(0),
												mPartSides(neither),
												mJoinUpdatesJoinField(sAllJoinsUpdateJoinField)
{
	assert(dbConnect::underConstruction());   
	dbConnect::sCurrentlyConstructing->attachRelation(this);
}


dbRelationship::~dbRelationship() 
{
	delete mlhs;
	delete mrhs;
}


dbRelationship& dbRelationship::names(const char* lhs, const char* rhs)
{
	assert(mlhs);
	assert(mlhs->mName.isEmpty());
	mlhs->mName = lhs;
		
	assert(mrhs);
	assert(mrhs->mName.isEmpty());
	mrhs->mName = rhs;
	return *this;
}


dbRelationship& dbRelationship::tables(dbTable* lhs, dbTable* rhs)
{
	assert(mlhs);
	assert(mlhs->mTable==0);
	mlhs->mTable = lhs;
		
	assert(mrhs);
	assert(mrhs->mTable==0);
	mrhs->mTable = rhs;
	return *this;
}


dbRelationship& dbRelationship::joinFields(dbField* lhs, dbField* rhs)
{
	assert(mlhs);
	assert(mlhs->mJoinField==0);
	mlhs->mJoinField = lhs;
	mlhs->mJoinToField = rhs;

	assert(mrhs);
	assert(mrhs->mJoinField==0);
	mrhs->mJoinField = rhs;
	mrhs->mJoinToField = lhs;
	
	if (mlhs->mTable==0) {
		assert(mrhs->mTable==0);
		mlhs->mTable = lhs->fieldTable();
		mrhs->mTable = rhs->fieldTable();
	}
	return *this;
}
 
 
void dbRelationship::lhs(dbTable* tbl, dbRelRefBase& linkFld, const char* name)
{
	assert(mlhs==0);    // should NEVER call this on a partly constructed relation 
	if (!mlhs) {
		mlhs = new dbRelHalf(tbl, linkFld, name);
		assert(mlhs);
	}
}


void dbRelationship::rhs (dbTable* tbl, dbRelRefBase& linkFld, const char* name)
{
	assert(mrhs==0);    // should NEVER call this on a partly constructed relation 
	if (!mrhs) {
		mrhs = new dbRelHalf(tbl, linkFld, name);
		assert(mrhs);
	}
} 



void dbRelationship::lhs(dbField* joinOver, dbRelRefBase& linkFld, const char* name)
{
	assert(mlhs==0);    // should NEVER call this on a partly constructed relation 
	if (!mlhs) {
		mlhs = new dbRelHalf(joinOver, linkFld, name);
		assert(mlhs);
	}
}


void dbRelationship::rhs (dbField* joinOver, dbRelRefBase& linkFld, const char* name)
{
	assert(mrhs==0);    // should NEVER call this on a partly constructed relation 
	if (!mrhs) {
		mrhs = new dbRelHalf(joinOver, linkFld, name);
		assert(mrhs);
	}
} 



void dbRelationship::joinsUpdateFields(bool joinFlag)
{
#ifdef OOF_Debug
// later have a check to stop them calling this after buildRelationship
#endif
	mJoinUpdatesJoinField = joinFlag;
}


void dbRelationship::buildRelationship()
{
	assert(mlhs);
	assert(mrhs);
// bit of a hack
// compensates for the way in which the lhs, rhs calls work
	if ((mlhs->mJoinField != 0) && (mlhs->mJoinToField == 0)){
// we have a relation based on joins, but because it was setup by lhs, rhs
// calls, the mJoinToFields weren't set (see joinFields() to understand why)
		mlhs->mJoinToField = mrhs->mJoinField;
		mrhs->mJoinToField = mlhs->mJoinField;
	}
	mlhs->mJoinUpdatesJoinField = mJoinUpdatesJoinField;
	mrhs->mJoinUpdatesJoinField = mJoinUpdatesJoinField;

	mlhs->buildRelationship(mrhs->mTable);
	mrhs->buildRelationship(mrhs->mTable);
}
 
 
void dbRelationship::describe(ostream& os)
{
	if (mName.isEmpty())
		os << "Un-named relation";
	else
		os << "Relation: " << mName;
	os << " consists of" << endl;
	if (mlhs)
		mlhs->describe(os);
	if (mrhs)
		mrhs->describe(os);		
	os << endl;
}


void dbRelationship::rhsOwnsLHSasParts()
{
	assert(mPartSides!=rhsIsPart);
	mPartSides = lhsIsPart;
}


void dbRelationship::lhsOwnsRHSasParts()
{
	assert(mPartSides!=lhsIsPart);
	mPartSides = rhsIsPart;
}
 
// -------------------------------------------------------
//            d b R e l H a l f
// -------------------------------------------------------
dbRelHalf::dbRelHalf(dbField* joinOver, dbRelRefBase& linkFldPtr, const char* name) :
					mTable(joinOver->fieldTable()),
					mInverseTable(0),
					mLinkfield(linkFldPtr),
					mName(name),
					mJoinField(joinOver),
					mJoinToField(0),
					mPartRelation(false)
					{};

void dbRelHalf::buildRelationship(dbTable* inverseTable)
{
	assert(mTable);
	if (mPartRelation)
		mTable->attachPartRelation(this);
	mInverseTable=inverseTable;
	mLinkfield.SetRelHalfByWhichWeNavigate(this);
}


void dbRelHalf::updateJoinField(dbTable* clonedTargetTable)
{
		dbField* targetField = clonedTargetTable->field(mJoinToField->fieldNumber());
		targetField->copyValueFrom(mJoinField);
}


void dbRelHalf::describe(ostream& os)
{
	assert(mTable);
	assert(mInverseTable);
	if (mName.isEmpty())
		os << "Un-named relation half";
	else
		os << "Relation half: " << mName;
	os << endl << "   links table " << mTable->tableName() << " to table: " << mInverseTable->tableName() << endl;
	if (mJoinField)
		os << endl << "   by dynamic join over field: " << mJoinField->fieldName() << endl;
}


// -------------------------------------------------------
//               d b R e l C h a i n
// -------------------------------------------------------
dbRelChain::~dbRelChain()
{ 
	delete mTablesList;
}


void dbRelChain::ConvertToRealChain()
{
	mTablesList = new OOF_DictRep(count(), 1);
	assert(mTablesList);
	dbTable* newTable;
   unsigned int numEntries = count();
	for (unsigned int i=0; i< numEntries; i++) {
		dbRelHalf* rel = (dbRelHalf* ) (operator[](i));  // safe downcast
		dbTable* destTable = rel->relatesToTable();
		newTable = destTable->cloneWithoutSelection();
	// NOT YET IMPLEMENTED - may have decision to make about following setSaveOptions depending on type of relation
		newTable->setSaveOption(dbTable::requireExplicitAndBuffer);
		mTablesList->Append(newTable);
	}
	mFinalTable = newTable;  // cache for lots of access calls
	newTable->assignRelChainToAllFields(this);
}


dbRelChain* dbRelChain::buildChain()
{
	assert(isStillJustACandidate());
	dbRelHalf* baseRel = (dbRelHalf* ) (operator[](0));  // safe downcast
	return baseRel->askTableToBuildValidChain(this);
}


void dbRelChain::validateContext()
{
	if (!mContextValid)
	{
		unloadRecord();
      unsigned int numEntries = count();
		for (unsigned int i=0; i< numEntries; i++) {
			dbRelHalf* rel = (dbRelHalf* ) (operator[](i));  // safe downcast
			rel->refreshContext((dbTable*) (mTablesList->item(i)) );   // safe downcast
		}
		mContextValid = true;
	}
}



dbField* dbRelChain::getFinalField(fieldNumT fieldNumber)
{
	return mFinalTable->field(fieldNumber);
}


void dbRelChain::newRecord() 
{
	if (!mContextValid) {
		unloadRecord();  //  should have been unloaded by now, but just in case
		mContextValid = true;
	}	
	mFinalTable->newRecord();
	propagateJoinCopy();
}


void dbRelChain::startRelated()
{
	validateContext();
	mFinalTable->start();
}


bool dbRelChain::moreRelated()
{
	validateContext();
	return mFinalTable->more();
}


void dbRelChain::nextRelated()
{
	mFinalTable->next();
}


unsigned long dbRelChain::countRelated()
{
	validateContext();
	return mFinalTable->count();
}


bool dbRelChain::gotoRelativeRecord(unsigned long recNo)
{
	validateContext();
	return mFinalTable->gotoRelativeRecord(recNo);
}


void dbRelChain::saveRecord()
{
	mFinalTable->saveRecord();
}


void dbRelChain::deleteRecord()
{
	validateContext();  // may not have been any field access to trigger a validateContext
	mFinalTable->deleteSelection();
}


void dbRelChain::propagateJoinCopy()
{
	unsigned int numEntries = count();
	for (unsigned int i=0; i< numEntries; i++) {
		dbRelHalf* rel = (dbRelHalf* ) (operator[](i));  // safe downcast
		if (rel->mJoinUpdatesJoinField)
			rel->updateJoinField((dbTable*) (mTablesList->item(i)) );   // safe downcast
	}
}


void dbRelChain::unloadRecord()
{
	mFinalTable->unloadRecord();
}


// -------------------------------------------------------
//                      d b C o n n e c t
// -------------------------------------------------------

void dbConnect::attachRelation(dbRelationship * theRel) 
{
	mRelations.append(theRel);
} 


// -------------------------------------------------------
//                      d b T a b l e
// -------------------------------------------------------
void dbTable::attachPartRelation(dbRelHalf * theHalf)  {
	mPartRelations.append(theHalf);
}



void dbTable::AddHalfRelation(dbRelRefBase& linkfieldPtr)
{
	if (sSavedRelationHalf) {
		new dbRelationship(sSavedRelationHalf, new dbRelHalf(this, linkfieldPtr));
		sSavedRelationHalf = 0;
	}
	else {
		sSavedRelationHalf = new dbRelHalf(this, linkfieldPtr);
		assert(sSavedRelationHalf);
	}
}


void dbTable::joinField(dbField& joinOn)
{
// NOT YET IMPLEMENTED
}


void dbTable::pathNamedAs(const char* name)
{
// NOT YET IMPLEMENTED
}

	
dbRelChain* dbTable::buildRelChain(dbRelChain* candidateChain)
{
	if (mRelChains) {
		dbRelChain* ret = mRelChains->match(candidateChain);
		if (ret) {
// NO - don't evaluate the relation at this point			ret->validateContext();  // in case changed!
			delete candidateChain;
			return ret;
		}
	}
	else {
		mRelChains = new OOF_relChainList;
		assert(mRelChains);
	}

	candidateChain->ConvertToRealChain();
	mRelChains->append(candidateChain);
	return candidateChain;
}



void dbTable::assignRelChainToAllFields(dbRelChain* pointsToField)
{
   unsigned int numFields = mFields.count();
	for (unsigned int i=0; i<numFields; i++)
	{
			dbField  *theField = (dbField  *) mFields[i];  // safe downcast
			theField->mRelationChain = pointsToField; 
	}	
}


// -------------------------------------------------------
//     O O F _ m i x R e l C h a i n E n d P o i n t
// -------------------------------------------------------

void OOF_mixRelChainEndPoint::appendRelationLink(dbRelHalf* relHalf)
{
	if (!sTransitoryRelChainLinks) {
		sTransitoryRelChainLinks = new dbRelChain;
		assert(sTransitoryRelChainLinks);
	}
	sTransitoryRelChainLinks->append(relHalf);
}


OOF_mixRelChainEndPoint::~OOF_mixRelChainEndPoint()
{
	delete mRelationChain;
}


// -------------------------------------------------------
//             O O F _ r e l C h a i n L i s t
// -------------------------------------------------------
dbRelChain* OOF_relChainList::match(dbRelChain* candidate)
{
	unsigned int numEntries = count();
	for (unsigned int i=0; i<numEntries; i++) {
		dbRelChain* theChain = (dbRelChain*) operator[](i);  // safe downcast
		if (theChain->allObjectsMatch(* (OOF_Dictionary*)candidate))
			return theChain;
	}
	return 0;
}


void OOF_relChainList::baseContextHasChanged()
{
	unsigned int numEntries = count();
	for (unsigned int i=0; i<numEntries; i++) {
		dbRelChain* theChain = (dbRelChain*) operator[](i);  // safe downcast
		theChain->baseContextHasChanged();
	}
}


void OOF_relChainList::propagateSave()
{
	unsigned int numEntries = count();
	for (unsigned int i=0; i<numEntries; i++) {
		dbRelChain* theChain = (dbRelChain*) operator[](i);  // safe downcast
		theChain->saveRecord();
	}
}


void OOF_relChainList::propagateDelete()
{
	unsigned int numEntries = count();
	for (unsigned int i=0; i<numEntries; i++) {
		dbRelChain* theChain = (dbRelChain*) operator[](i);  // safe downcast
		theChain->deleteRecord();
	}
}

void OOF_relChainList::propagateUnload()
{
	unsigned int numEntries = count();
	for (unsigned int i=0; i<numEntries; i++) {
		dbRelChain* theChain = (dbRelChain*) operator[](i);  // safe downcast
		theChain->unloadRecord();
	}
}

// -------------------------------------------------------
//              d b R e l R e f B a s e 
// -------------------------------------------------------
void dbRelRefBase::BuildRelChainToTable() 
{
	LogTransition();	// fake the effect of the further ->field that would occur for field of this ref
	ConsumePossibleRelationChain();
	assert(mRelationChain);  // if we dn't have a chain, then why are we here at a ref?
	mRelationChain = mRelationChain->buildChain();
}


bool dbRelRefBase::gotoRelativeRecord(unsigned long recNo)
{
	if (!mRelationChain)
		BuildRelChainToTable();
	return mRelationChain->gotoRelativeRecord(recNo);
}


